#!/usr/bin/env python
# -*- coding: utf-8 -*-


# __module__ set in class definitions so they show as e.g. "ReactionError" instead of "schem.exceptions.ReactionError"
# in exception messages


class SolutionImportError(Exception):
    """Raised when a solution string cannot be successfully loaded into a level."""
    __module__ = Exception.__module__


class SolutionRunError(RuntimeError):
    """Raised when a solution encounters an error while being run."""
    __module__ = Exception.__module__


class ReactionError(SolutionRunError):
    """Raised if there is a molecule collision or a molecule is pulled apart in a reactor."""
    __module__ = Exception.__module__


class PauseException(SolutionRunError):
    """Raised when a Pause command is encountered by a waldo."""
    __module__ = Exception.__module__


class InvalidOutputError(SolutionRunError):
    """Raised when an invalid molecule is passed to a level output zone."""
    __module__ = Exception.__module__


class ScoreError(Exception):
    """Raised during validations if a solution's expected score does not match its actual score."""
    __module__ = Exception.__module__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import namedtuple
from enum import Enum

from .grid import *


class InstructionType(Enum):
    '''Represents the various types of SpaceChem non-arrow instruction.'''
    START = 'S'
    INPUT = 'i'
    OUTPUT = 'o'
    GRAB_DROP = 'g-d'
    GRAB = 'grb'
    DROP = 'drp'
    ROTATE = 'r'  # ⭯ ⭮
    SYNC = 'sy'  # ⌛
    BOND_PLUS = 'b+'  # ⊕
    BOND_MINUS = 'b-'
    SENSE = '?'
    FLIP_FLOP = 'ff'  # F
    FUSE = 'fus'  # ⛯☢️
    SPLIT = 'spl'  # ☢️
    SWAP = 'swp'  # ⇄
    PAUSE = 'P'  # ⏸

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.value[0]


class Instruction(namedtuple('Instruction', ('type', 'direction', 'target_idx'),
                             # Default direction and target_idx if unneeded
                             defaults=(None, None))):
    '''Represents a non-arrow SpaceChem instruction and any associated properties.
    Direction is used by rotates, flip-flops and sensor cmds. Target index represents either the zone idx for an input
    or output instruction, or an element's atomic # in the case of a sensor cmd.
    '''
    __slots__ = ()

    STR_MAP = {InstructionType.START: 'S',
               InstructionType.INPUT: 'i',
               InstructionType.OUTPUT: 'o',
               InstructionType.GRAB_DROP: '&',
               InstructionType.GRAB: 'G',
               InstructionType.DROP: 'D',
               InstructionType.ROTATE: {Direction.CLOCKWISE: '⭮', Direction.COUNTER_CLOCKWISE: '⭯'},
               InstructionType.SYNC: '⌛',
               InstructionType.BOND_PLUS: '+',
               InstructionType.BOND_MINUS: '-',
               InstructionType.SENSE: '?',
               InstructionType.FLIP_FLOP: 'F',  # △ ▷ ▽ ◁ ▲ ▶ ◀ ▼
               InstructionType.FUSE: '⛯',  # ☢️ ?
               InstructionType.SPLIT: '☢',
               InstructionType.SWAP: '⇔',  # ⇄ ?
               InstructionType.PAUSE: 'P'}

    def __repr__(self):
        return (f'Instruction({repr(self.type)}'
                + (f', target_idx={self.target_idx}' if self.target_idx is not None else '')
                + (f', direction={repr(self.direction)}' if self.direction is not None else '')
                + ')')

    def __str__(self):
        s = self.STR_MAP[self.type]
        return s if isinstance(s, str) else s[self.direction]

    # TODO: Define from_export_str to simplify Reactor.from_export_str

    def export_str(self, waldo_idx, posn):
        # Handle instructions with names/sub-types that I deliberately differed from SC's implementation
        special_case_dict = {
            InstructionType.GRAB_DROP: ('grab', 0),
            InstructionType.GRAB: ('grab', 1),
            InstructionType.DROP: ('grab', 2),
            InstructionType.ROTATE: ('rotate', 0 if self.direction == Direction.CLOCKWISE else 1),
            InstructionType.BOND_PLUS: ('bond', 0),
            InstructionType.BOND_MINUS: ('bond', 1),
            InstructionType.SENSE: ('sensor', 0),
            InstructionType.FLIP_FLOP: ('toggle', 0)}
        if self.type in special_case_dict:
            name, sub_type = special_case_dict[self.type]
        else:
            name, sub_type = self.type.name.lower(), self.target_idx if self.target_idx is not None else 0

        # Direction stored in degrees with UP = -90, LEFT = 180
        if self.direction is None or self.type == InstructionType.ROTATE:
            dirn_degrees = -1
        else:
            dirn_degrees = ((self.direction.value - 1) * 90)

        waldo_int = 128 if waldo_idx == 0 else 32  # Instructions are 32 for blue, 128 for red
        target_element = self.target_idx if self.type == InstructionType.SENSE else 0

        # Dunno what second last field does
        return f"MEMBER:'instr-{name}',{dirn_degrees},{sub_type},{waldo_int},{posn.col},{posn.row},0,{target_element}"


class Waldo:
    __slots__ = 'idx', 'instr_map', 'flipflop_states', 'position', 'direction', 'molecule', 'is_stalled', 'is_rotating'

    # Map of good unicode characters for displaying a waldo's paths
    dirns_to_char = {
        # https://www.compart.com/en/unicode/charsets/ISO_10367-box (light box chars): ┌,─,┐,└,│,┘,┼,┴,┬,┤,├,╴,╶,╵,╷
        frozenset((UP,)): '│',    # ╵ sadly doesn't work in common fonts
        frozenset((DOWN,)): '│',  # ╷
        frozenset((LEFT,)): '─',  # ╴
        frozenset((RIGHT,)): '─', # ╶
        frozenset((UP, RIGHT)): '└',
        frozenset((UP, DOWN)): '│',
        frozenset((UP, LEFT)): '┘',
        frozenset((RIGHT, DOWN)): '┌',
        frozenset((RIGHT, LEFT)): '─',
        frozenset((DOWN, LEFT)): '┐',
        frozenset((UP, RIGHT, DOWN)): '├',
        frozenset((UP, RIGHT, LEFT)): '┴',
        frozenset((UP, DOWN, LEFT)): '┤',
        frozenset((RIGHT, DOWN, LEFT)): '┬',
        frozenset((UP, RIGHT, DOWN, LEFT)): '┼'}

    def __init__(self, idx, instr_map):
        self.idx = idx
        self.instr_map = instr_map  # Position map of tuples containing arrow (direction) and 'command' (non-arrow) instructions
        self.flipflop_states = {}
        self.position = None  # Dummy init so we can check for duplicate Start
        for posn, (_, cmd) in instr_map.items():
            if cmd is not None:
                if cmd.type == InstructionType.FLIP_FLOP:
                    self.flipflop_states[posn] = False
                elif cmd.type == InstructionType.START:
                    assert self.position is None, "Duplicate waldo Start instruction"
                    self.position = posn
                    self.direction = cmd.direction
        assert self.position is not None, "Missing waldo Start instruction"

        self.molecule = None
        self.is_stalled = False  # Updated if the waldo is currently stalled on a sync, output, wall, etc.
        self.is_rotating = False  # Used to distinguish the first vs second cycle on a rotate command

    def __hash__(self):
        '''A waldo's run state is uniquely identified by its position, direction,
        whether it's holding a molecule, whether it's rotating, and its flip-flop states.
        '''
        return hash((self.position, self.direction, self.molecule is None, self.is_rotating,
                     tuple(self.flipflop_states.values())))

    def __repr__(self):
        return (f'Waldo({self.idx}, pos={self.position}, dir={self.direction}, is_stalled={self.is_stalled}'
                + f', instr_map={self.instr_map})')

    def __len__(self):
        return sum(1 for cell_instrs in self.instr_map.values()
                   for instr in cell_instrs
                   if (instr is not None
                       and not (isinstance(instr, Instruction) and instr.type == InstructionType.START)))

    def cur_cmd(self):
        '''Return a waldo's current 'command', i.e. non-arrow instruction, or None.'''
        return None if self.position not in self.instr_map else self.instr_map[self.position][1]

    def export_str(self):
        '''Represent this waldo's instructions in solution export string format.'''
        lines = []
        for posn, (arrow, instr) in self.instr_map.items():
            if arrow is not None:
                waldo_int = 64 if self.idx == 0 else 16
                dirn_degrees = (arrow.value - 1) * 90  # TODO: Should probably stop using .value for IntEnum's?
                lines.append(f"MEMBER:'instr-arrow',{dirn_degrees},0,{waldo_int},{posn.col},{posn.row},0,0")

            if instr is not None:
                lines.append(instr.export_str(waldo_idx=self.idx, posn=posn))

        return '\n'.join(lines)

    def trace_path(self, num_cols=10, num_rows=8):
        """Return a dict of position: directions representing paths this waldo visually traces.
        Useful for representing the waldo visually and for calculating e.g. waldopath.
        """
        def is_valid_posn(posn):
            return 0 <= posn.col < num_cols and 0 <= posn.row < num_rows

        branching_instr_types = set((InstructionType.SENSE, InstructionType.FLIP_FLOP))

        # Override the start direction with any arrow since unlike other directional commands it can't branch
        start_posn, start_dirn = next((posn, arrow_dirn if arrow_dirn is not None else cmd.direction)
                                      for posn, (arrow_dirn, cmd) in self.instr_map.items()
                                      if cmd.type == InstructionType.START)
        visited_posn_dirns = {}  # posn: directions to catch when we're looping
        traced_posn_dirns = {}  # posn: directions that were visually traced (separate from visited since
                                # a cell that is passed through in only one direction may be traced in both)
        unexplored_branches_stack = [(start_posn, start_dirn)]
        while unexplored_branches_stack:
            cur_posn, incoming_dirn = unexplored_branches_stack.pop()

            # Trace the path coming into this cell
            if cur_posn not in traced_posn_dirns:
                traced_posn_dirns[cur_posn] = set()
            traced_posn_dirns[cur_posn].add(incoming_dirn.opposite())

            # Check the current cell for an arrow and/or branching instruction
            arrow_dirn, cmd = self.instr_map[cur_posn] if cur_posn in self.instr_map else (None, None)

            # Arrows update the direction of the current branch but don't create a new one
            outgoing_dirn = arrow_dirn if arrow_dirn is not None else incoming_dirn

            # Check the current position/direction against the visit map. We do this after evaluating the arrow to
            # reduce excess visits (since the original direction of a waldo never matters to its outgoing path if an
            # arrow is present, unlike with branching commands)
            if cur_posn not in visited_posn_dirns:
                visited_posn_dirns[cur_posn] = set()
            elif outgoing_dirn in visited_posn_dirns[cur_posn]:
                # We've already explored this cell in the current direction and must have already added any branches
                # starting from this cell, so end this branch
                continue

            traced_posn_dirns[cur_posn].add(outgoing_dirn)
            visited_posn_dirns[cur_posn].add(outgoing_dirn)

            # Add any new branch
            if cmd is not None and cmd.type in branching_instr_types:
                traced_posn_dirns[cur_posn].add(cmd.direction)
                next_branch_posn = cur_posn + cmd.direction
                if is_valid_posn(next_branch_posn):
                    unexplored_branches_stack.append((next_branch_posn, cmd.direction))

            # Put the current branch back on top of the stack
            next_posn = cur_posn + outgoing_dirn
            if is_valid_posn(next_posn):
                unexplored_branches_stack.append((next_posn, outgoing_dirn))

        return traced_posn_dirns

    def reset(self):
        for posn, (_, cmd) in self.instr_map.items():
            if cmd is not None and cmd.type == InstructionType.START:
                self.position = posn
                self.direction = cmd.direction
                break

        self.molecule = None
        self.is_stalled = False
        self.is_rotating = False

        for k in self.flipflop_states:
            self.flipflop_states[k] = False

        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional

from .level import Level
from .levels import levels as built_in_levels, defense_names, resnet_ids
from .exceptions import SolutionImportError, ScoreError
from .precognition import is_precognitive
from .solution import Solution, DebugOptions


def load_solution(soln_str: str, level_codes=None, _return_resnet_id=False) -> Solution:
    """Given a solution export string, return a Solution object loaded into the level with matching name.

    If multiple official or given levels match the level name in the solution's metadata, returns the first Solution
    object that can be successfully loaded from them. Note that in the case of official levels, there are currently no
    levels sharing a name that have compatible features, so this is guaranteed to load the correct level if possible.
    If multiple custom levels of the same name are provided, no guarantee is made that the solution doesn't load into
    multiple of them; only the first successful load is returned.

    If level_codes is provided, only the given level(s) are checked (official levels are ignored).

    Note: To attempt to load a solution into a level with mis-matched name, use the Solution constructor directly.
    """
    level_name = Solution.parse_metadata(soln_str)[0]

    # If no levels were provided, fetch the matching official level(s)
    if level_codes is None:
        # Determine the built-in game level to run the solution against based on the level name in its metadata
        if level_name in built_in_levels:
            level_codes = []
            matching_resnet_ids = []
            if isinstance(built_in_levels[level_name], str):
                level_codes.append(built_in_levels[level_name])
                matching_resnet_ids.append(resnet_ids[level_name] if level_name in resnet_ids else None)
            else:
                level_codes.extend(built_in_levels[level_name])
                matching_resnet_ids.extend(resnet_ids[level_name] if level_name in resnet_ids
                                           else [None for _ in range(len(built_in_levels[level_name]))])
        elif level_name in defense_names:
            raise NotImplementedError("Defense levels unsupported")
        else:
            raise Exception(f"No known level `{level_name}`")
    else:
        # Return None for the resnet ID if not defaulting to official levels
        matching_resnet_ids = [None] * len(level_codes)

    exceptions = []
    for level_code, resnet_id in zip(level_codes, matching_resnet_ids):
        level = Level(level_code)
        if level.name == level_name:
            try:
                solution = Solution(level=level, soln_export_str=soln_str)
                return solution if not _return_resnet_id else (solution, resnet_id)
            except SolutionImportError as e:
                exceptions.append(e)

    raise exceptions[0] if exceptions else Exception(f"No level `{level_name}` provided")


def run(soln_str: str, level_code=None, level_codes=None, max_cycles=None, return_json=False,
        check_precog=False, max_precog_check_cycles=None,
        verbose=False, stderr_on_precog=False,
        debug: Optional[DebugOptions] = False,
        _validate_expected_score=False):
    """Wrapper for Solution.run which identifies the level to run in as needed, based on the solution metadata.

    Run the given solution, and return the (cycles, reactors, symbols) Score obtained. Raise an error if the solution
    cannot run to completion or exceeds max_cycles.

    Args:
        soln_str: Solution string as exported by SpaceChem Community Edition.
        level_code / level_codes: Puzzle export string or strings (only one or the other arg may be provided).
            If multiple levels are given, solution will be run against all with titles matching the level title in its
            metadata, and an exception raised if none match. If only one level is provided, the solution will be run
            against it, irrespective of title.

            If level_code(s) is not provided, the level name from the solution metadata is used to search for a matching
            built-in game level(s).

            When multiple built-in or provided level titles match the solution's metadata, the solution will be run
            against all in turn until one matches the expected score. If no level matches the expected score, the first
            successful run's score will be returned. If the solution does not complete any of the levels, an exception
            is raised.
            Note that while some built-in game levels have duplicate titles, currently all of these have conflicting
            level features (bonders etc.), and therefore a solution can't successfully load into / solve multiple
            built-in levels.
        max_cycles: Maximum cycle count to run to. Default 1.1x the expected cycle count in the solution metadata,
            or 1,000,000 cycles if no expected score (use math.inf if you don't fear infinite loop solutions).
        return_json: If True, instead of a Score return a dict including the usual score fields, but also the level
            title, ResearchNet volume-issue-puzzle tuple (None if not a ResearchNet level), and solution author/title.

            Additionally, raise an exception only if the solution cannot be imported into any level; if the solution
            encounters a reaction error or exceeds max_cycles, just return None in the dict's 'cycles' field and
            return the metadata of the first level that the solution could be successfully imported into.

            Default False.
        check_precog: If True, do additional runs on the solution to check if it fits the current community definition
            of a precognitive solution. Currently only useful if return_json is present, in which case a 'precog' field
            will be included. See `schem.precognition.is_precognitive` for more info and a direct API.
        max_precog_check_cycles: The maximum total cycle count that may be used by all precognition-check runs; if this
            value is exceeded before sufficient confidence in an answer is obtained, a TimeoutError is raised, or in the
            case of return_json, the 'precog' field is set to None.
            Default 2,000,000 cycles (this is sufficient for basically any sub-10k solution).
        verbose: If True, print warnings if there is not exactly one level with title matching the solution metadata.
                 Default False.
        stderr_on_precog: If True, when a solution is precognitive, print an explanation of why to STDERR.
                          Can be enabled independently of `verbose`. Default False.
        debug: Print an updating view of the solution while running; see DebugOptions. Default False.
    """
    assert level_code is None or level_codes is None, "Only one of level_code or level_codes may be specified"
    level_name, author, expected_score, soln_name = Solution.parse_metadata(soln_str)
    soln_descr = Solution.describe(level_name, author, expected_score, soln_name)

    if _validate_expected_score:
        if expected_score is None:
            raise ValueError("Validation requires a valid expected score in the first solution line (currently 0-0-0);"
                             + " please update it or use run() without validation instead.")

        if max_cycles is not None:
            if expected_score.cycles > max_cycles:
                raise ValueError(f"{soln_descr}: Cannot validate; expected cycles > max cycles ({max_cycles})")

            # If validating, limit the max cycles based on the expected score, to save time
            max_cycles = min(max_cycles, int(expected_score.cycles * 1.1))

    # Convert level_code convenience arg to same format as level_codes
    if level_code is not None:
        level_codes = [level_code]

    # Load the solution, skipping strict level name matching if the singular level_code was given.
    if level_code is None:
        solution, resnet_id = load_solution(soln_str, level_codes=level_codes, _return_resnet_id=True)
    else:
        solution, resnet_id = Solution(level=Level(level_code), soln_export_str=soln_str), None

    # Exit early if we're validating and the reactor or symbol count doesn't match that expected
    if _validate_expected_score:
        reactors = len(list(solution.reactors))
        if reactors != expected_score.reactors:
            raise ScoreError(f"{soln_descr}: Expected {expected_score.reactors} reactors but got {reactors}.")

        symbols = solution.symbols
        if symbols != expected_score.symbols:
            raise ScoreError(f"{soln_descr}: Expected {expected_score.symbols} symbols but got {symbols}.")

    score = cycles, reactors, symbols = solution.run(max_cycles=max_cycles, debug=debug)

    # Skip precog check if expected score validation failed
    if _validate_expected_score and score != expected_score:
        raise ScoreError(f"{soln_descr}: Expected score {expected_score} but got {score}")

    run_data = {'level_name': solution.level.name,
                'resnet_id': resnet_id,
                'cycles': cycles,
                'reactors': reactors,
                'symbols': symbols,
                'author': solution.author,
                'solution_name': solution.name}

    if check_precog:
        try:
            run_data['precog'] = is_precognitive(solution, max_cycles=max_cycles,
                                                 max_total_cycles=max_precog_check_cycles,
                                                 just_run_cycle_count=score.cycles,
                                                 verbose=verbose,
                                                 stderr_on_precog=stderr_on_precog)
        except TimeoutError:
            # If using --json mode, store None for the field instead of propagating any timeout error
            if return_json:
                run_data['precog'] = None
            else:
                raise

    if verbose and _validate_expected_score:
        print(f"Validated {soln_descr}")

    return run_data if return_json else score


def validate(soln_str: str, level_code=None, level_codes=None, max_cycles=None, return_json=False, check_precog=False,
             max_precog_check_cycles=None, verbose=False, stderr_on_precog=False,
             debug: Optional[DebugOptions] = False):
    """Sibling of Solution.validate which identifies the level to run in as needed, based on the solution metadata.

    Run the given solution, and raise an exception if the score does not match that indicated in the solution metadata.

    Args:
        soln_str: Solution string as exported by SpaceChem Community Edition.
        level_code / level_codes: Puzzle export string or strings (only one or the other arg may be provided).
            If multiple levels are given, solution will be run against all with titles matching the level title in its
            metadata, and an exception raised if none match. If only one level is provided, the solution will be run
            against it, irrespective of title.

            If level_code(s) is not provided, the level name from the solution metadata is used to search for a matching
            built-in game level(s).

            When multiple built-in or provided level titles match the solution's metadata, the solution will be run
            against all in turn until one matches the expected score, or raise an error if none do.
            Note that while some built-in game levels have duplicate titles, currently all of these have conflicting
            level features (bonders etc.), and therefore a solution can't successfully load into / solve multiple
            built-in levels.
        max_cycles: Maximum cycle count to run to. Default 1.1x the expected cycle count in the solution metadata.
        return_json: If True, instead of a Score return a dict including the usual score fields, but also the level
            title, ResearchNet volume-issue-puzzle tuple (None if not a ResearchNet level), and solution author/title.

            Additionally, raise an exception only if the solution cannot be imported into any level; if the solution
            encounters a reaction error or exceeds max_cycles, just return None in the dict's 'cycles' field and
            return the metadata of the first level that the solution could be successfully imported into.

            Default False.
        check_precog: If True, do additional runs on the solution to check if it fits the current community definition
            of a precognitive solution. Currently only useful if return_json is present, in which case a 'precog' field
            will be included. See `schem.precognition.is_precognitive` for more info and a direct API.
        max_precog_check_cycles: The maximum total cycle count that may be used by all precognition-check runs; if this
            value is exceeded before sufficient confidence in an answer is obtained, a TimeoutError is raised, or in the
            case of return_json, the 'precog' field is set to None.
            Default 2,000,000 cycles (this is sufficient for basically any sub-10k solution).
        verbose: If True, print warnings if there is not exactly one level with title matching the solution metadata.
                 Default False.
        stderr_on_precog: If True, when a solution is precognitive, print an explanation of why to STDERR.
                          Can be enabled independently of `verbose`. Default False.
        debug: Print an updating view of the solution while running; see DebugOptions. Default False.
    """
    ret_val = run(soln_str, level_code=level_code, level_codes=level_codes, max_cycles=max_cycles,
                  return_json=return_json, check_precog=check_precog,
                  max_precog_check_cycles=max_precog_check_cycles,
                  verbose=verbose, stderr_on_precog=stderr_on_precog, debug=debug,
                  _validate_expected_score=True)

    if return_json:
        return ret_val



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from itertools import product

MAX_TERRAIN_INT = 7

# Terrain obstacles and default component positions. The positions of obstacles within the level is uniquely identified by the terrain type for
# all ResearchNet and custom levels, but each main game level has a unique terrain layout.
# This dict will store, for each type indexed by either its level name (main game) or its
# integer terrain type (ResearchNet/custom), the following:
# * Set containing all co-ordinates obstructed by terrain obstacles (rocks, etc.)
# * lists of where inputs, outputs, and other overworld buildings will be spawned if present

# itertools.product used to somewhat more compactly specify the co-ordinates in obstacles
terrains = {
    'research':
        {'obstructed': {},
         'input-zones': (('research-input', (0, 1)),  # actually these have empty type in SC but fuck that
                         ('research-input', (0, 2))),
         'output-zones': (('research-output', (7, 1)),
                          ('research-output', (7, 2)),)},
    0: {  # Sernimir IV
        'obstructed': {
            # Top-left rock
            *product(range(0, 10), (0,)),
            *product(range(0, 9), (1,)),
            *product(range(0, 8), (2,)),
            *product(range(2, 7), (3,)),
            # Mid-left rock
            *product(range(4, 6), (8,)),
            *product(range(3, 7), (9,)),
            *product(range(2, 8), range(10, 13)),
            *product(range(5, 7), (13,)),
            # Bottom rock
            *product(range(18, 21), (16,)),
            *product(range(14, 22), range(17, 19)),
            *product(range(15, 22), (19,)),
            *product(range(17, 21), (20,)),
            # Top-right rock
            *product(range(26, 32), range(0, 2)),
            *product(range(27, 32), (2,)),
            *product(range(28, 32), (3,)),
            *product(range(29, 32), (4,)),
            # Mid-right rock
            *product(range(30, 32), range(13, 19)),
            *product((29,), range(14, 18)),
        },
        'random-input-zones': (('drag-arbitrary-input', (1, 5)),),
        'fixed-input-zones': (('drag-arbitrary-input', (3, 13)),
                              ('drag-arbitrary-input', (3, 18))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 7)),
                         ('output', (23, 12))),
        'recycler': ('drag-recycler', (24, 17))},
    1: {  # Danopth
        'obstructed': {
            # Top-left rock/pit
            *product(range(3, 5), (0,)),
            *product(range(2, 6), (1,)),
            *product(range(1, 12), range(2, 4)),
            *product(range(1, 11), (4,)),
            *product(range(3, 5), (5,)),
            # Top-middle rock
            *product(range(19, 23), range(0, 2)),
            # Bottom-left rock
            *product(range(0, 2), range(15, 19)),
            *product((0,), range(19, 21)),
            *product(range(0, 7), (21,)),
            # Bottom-right rock
            *product((31,), (15,)),
            *product(range(29, 32), range(16, 19)),
            *product(range(30, 32), (19,)),
        },
        'random-input-zones': (('drag-arbitrary-input', (1, 5)),),
        'fixed-input-zones': (('drag-arbitrary-input', (3, 13)),
                              ('drag-arbitrary-input', (3, 18))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 7)),
                         ('output', (23, 12))),
        'recycler': ('drag-recycler', (24, 17))},
    2: {  # Alkonost
        'obstructed': {
            # Left tree column
            *product((0,), range(0, 22)),
            # Top-left trees
            *product(range(1, 6), range(0, 3)),
            *product(range(1, 5), (3,)),
            # Bottom-left trees
            *product((1,), range(14, 18)),
            *product((2,), range(15, 17)),
            *product((1,), range(20, 22)),
            # Middle trees
            (14, 10),
            *product(range(12, 16), (11,)),
            *product(range(9, 16), range(12, 15)),
            *product(range(10, 16), (15,)),
            # Top-mid trees
            *product(range(13, 19), (0,)),
            *product(range(14, 18), (1,)),
            # Right trees
            *product((29,), range(4, 6)),
            *product((30,), range(2, 7)),
            *product((31,), range(2, 10)),
        },
        'random-input-zones': (('drag-arbitrary-input', (1, 5)),),
        'fixed-input-zones': (('drag-arbitrary-input', (3, 13)),
                              ('drag-arbitrary-input', (3, 18))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 7)),
                         ('output', (23, 12))),
        'recycler': ('drag-recycler', (24, 17))},
    3: {  # Sikutar
        'obstructed': {
            # Left ravine
            *product((0,), range(6, 8)),
            *product(range(0, 4), (8,)),
            *product(range(7, 9), (8,)),
            *product(range(0, 9), range(9, 12)),
            *product(range(0, 2), (12,)),
            # Middle snow
            *product(range(13, 16), range(7, 10)),
            # Top-right water
            *product(range(24, 32), (0,)),
            *product(range(26, 32), (0,)),
            # Right rocks
            (31, 11),
            *product(range(30, 32), range(12, 15)),
        },
        'random-input-zones': (('drag-arbitrary-input', (1, 5)),),
        'fixed-input-zones': (('drag-arbitrary-input', (3, 13)),
                              ('drag-arbitrary-input', (3, 18))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 7)),
                         ('output', (23, 12))),
        'recycler': ('drag-recycler', (24, 17))},
    4: {  # Hephaestus IV
        'obstructed': {
            # Bottom-left lava
            *product((0,), range(14, 20)),
            *product((1,), range(15, 19)),
            # Middle lava
            *product(range(10, 13), range(4, 8)),
        },
        'random-input-zones': (('drag-arbitrary-input', (1, 5)),),
        'fixed-input-zones': (('drag-arbitrary-input', (3, 13)),
                              ('drag-arbitrary-input', (3, 18))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 7)),
                         ('output', (23, 12))),
        'recycler': ('drag-recycler', (24, 17))},
    5: {  # Flidais
        'obstructed': {
            # Bottom-left creeper
            *product(range(0, 2), range(16, 18)),
            # Middle creeper
            *product(range(10, 17), range(7, 10)),
            *product(range(10, 16), (10,)),
            # Right creeper
            *product(range(26, 29), (11,)),
            *product((26,), range(12, 14)),
            *product(range(27, 32), range(12, 16)),
            (31, 16),
        },
        'random-input-zones': (('drag-arbitrary-input', (1, 5)),),
        'fixed-input-zones': (('drag-arbitrary-input', (3, 13)),
                              ('drag-arbitrary-input', (3, 18))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 7)),
                         ('output', (23, 12))),
        'recycler': ('drag-recycler', (24, 17))},
    6: {  # 63 Corvi
        'obstructed': {
            # Top wall
            *product(range(0, 32), (0,)),
            *product(range(0, 8), (1,)),
            # Raised platform's edges
            *product(range(9, 11), range(1, 8)),
            *product(range(9, 32), range(8, 10)),
            # Right end of platform
            *product(range(28, 32), range(1, 8)),
        }},
    7: {  # Sandbox
        'obstructed': {
            # Top-left rock
            *product(range(10), (0,)),
            *product(range(9), (1,)),
            *product(range(8), (2,)),
            *product(range(6), (3,)),
            (0, 4),
            # Top-right rock
            *product(range(25, 32), (0,)),
            *product(range(26, 32), (1,)),
            *product(range(27, 32), (2,)),
            *product(range(28, 32), (3,)),
            *product(range(29, 32), (4,)),
            # Bottom-left rock
            *product((0,), range(19, 21)),
        },
        'random-input-zones': (('drag-arbitrary-input', (1, 4)),),
        'programmed-input': ('drag-programmed-input', (1, 13)),
        'fixed-input-zones': (('drag-arbitrary-input', (2, 18)),),
        'recycler': ('drag-recycler', (27, 17))},
    # TODO: I only did the most central terrain obstacles on my first pass through the main game; add the edge bits
    "An Introduction to Pipelines": {
        'obstructed': {
            # Middle rock
            *product(range(18, 21), (16,)),
            *product(range(14, 22), range(17, 19)),
            *product(range(15, 22), (19,)),
            *product(range(17, 21), (20,)),
        },
        'fixed-input-zones': (('drag-silo-input', (9, 4)),
                              ('drag-silo-input', (8, 12))),
        'output-zones': (('output', (23, 6)),
                         ('output', (23, 12)),)},
    "There's Something in the Fishcake": {
        'obstructed': {
            # Top-left rock
            *product(range(4, 6), (1,)),
            *product(range(3, 7), (2,)),
            *product(range(2, 8), range(3, 6)),
            *product(range(3, 7), (6,)),
            # Middle rock
            *product(range(17, 21), (9,)),
            *product(range(15, 22), (10,)),
            *product(range(14, 22), range(11, 13)),
            *product(range(16, 22), (13,)),
            # Wreckage
            *product(range(24, 27), range(18, 20)),
            # TODO: Edge bits
        },
        'fixed-input-zones': (('drag-silo-input', (6, 8)),
                              ('drag-silo-input', (7, 16))),
        'output-zones': (('output', (23, 6)),)},
    "Sleepless on Sernimir IV": {
        'obstructed': {
            # Top rock
            *product(range(17, 21), (2,)),
            *product(range(15, 22), (3,)),
            *product(range(14, 22), range(4, 6)),
            *product(range(15, 22), (6,)),
        },
        'fixed-input-zones': (('drag-silo-input', (4, 2)),
                              ('drag-silo-input', (5, 9)),
                              ('drag-silo-input', (6, 16))),
        'output-zones': (('output', (23, 8)),)},
    "Settling into the Routine": {
        'obstructed': {},
        'fixed-input-zones': (('drag-oceanic-input', (8, 3)),),
        'output-zones': (('output', (23, 12)),
                         ('output', (23, 18)))},
    "Nothing Works": {
        'obstructed': {
            # Rock
            *product(range(20, 23), (9,)),
            *product(range(19, 23), range(10, 12)),
            *product(range(20, 23), (12,)),
        },
        'fixed-input-zones': (('drag-oceanic-input', (8, 10)),
                              ('drag-atmospheric-input', (5, 19))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 18)))},
    "Challenge: In-Place Swap": {
        'obstructed': {
            # Bottom-left stuff
            *product(range(3, 6), (17,)),
            *product(range(0, 7), range(18, 20)),
            *product(range(0, 8), (20,)),
            *product(range(0, 9), (21,)),
            # Bottom-middle rock
            *product(range(13, 16), (18,)),
            *product(range(12, 17), (19,)),
            *product(range(11, 19), (20,)),
            *product(range(10, 22), (21,)),
        },
        'fixed-input-zones': (('drag-atmospheric-input', (5, 4)),
                              ('drag-oceanic-input', (1, 16))),
        'output-zones': (('output', (23, 12)),
                          ('output', (23, 18),))},
    "No Ordinary Headache": {
        'obstructed': {
            # Middle trees
            *product(range(12, 15), (3,)),
            *product(range(11, 16), (4,)),
            *product(range(9, 16), range(5, 8)),
            *product(range(10, 16), (8,)),
        },
        'random-input-zones': (('drag-atmospheric-input', (8, 18)),),
        'output-zones': (('output', (23, 2)),),
        'recycler': ('drag-recycler', (25, 14))},
    "No Thanks Necessary": {
        'obstructed': {},
        'random-input-zones': (('drag-atmospheric-input', (4, 5)),
                               ('drag-atmospheric-input', (7, 18)),),
        'output-zones': (('output', (22, 7)),
                         ('output', (21, 17)),),
        'recycler': ('drag-recycler', (26, 1))},
    "Challenge: Going Green": {
        'obstructed': {
            # Middle trees
            *product(range(12, 15), (10,)),
            *product(range(10, 16), range(11, 16)),
            *product((9,), range(12, 15)),
        },
        'random-input-zones': (('drag-powerplant-input', (-7, 4)),),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 18)))},
    "Falling": {
        'obstructed': {
            # Top-left water
            *product(range(0, 2), (4,)),
            *product(range(0, 3), (5,)),
            *product(range(0, 9), range(6, 10)),
            *product(range(0, 2), (10,)),
            # Middle ice
            *product(range(13, 16), range(5, 8)),
        },
        'fixed-input-zones': (('drag-oceanic-input', (4, 4)),
                              ('drag-silo-input', (0, 15))),
        'output-zones': (('output', (23, 7)),
                         ('output', (23, 12)),
                         ('output', (23, 17))),
        'recycler': ('drag-recycler', (26, 1))},
    "Challenge: Applied Fusion": {
        'obstructed': {
            # Top-left water
            # You can tell Zach did these by hand since the same ravine as in falling is traced slightly differently
            *product(range(0, 3), range(11, 13)),
            *product(range(0, 9), range(13, 17)),
            *product(range(0, 2), (17,)),
            # Middle ice
            *product(range(13, 16), range(12, 8)),
        },
        'fixed-input-zones': (('drag-oceanic-input', (4, 11)),),
        'output-zones': (('output', (23, 18)),),
        'recycler': ('drag-recycler', (26, 1))},
    "Molecular Foundry": {
        'obstructed': {
            # Middle rock
            *product(range(10, 13), range(11, 15)),
            # Wreckage
            *product(range(2, 5), range(18, 20)),
        },
        'random-input-zones': (('drag-mining-input', (1, 7)),),
        'output-zones': (('output', (23, 18)),),
        'recycler': ('drag-recycler', (22, 3))},
    "Gas Works Park": {
        'obstructed': {
            # Middle rock
            *product(range(10, 13), range(4, 8)),
        },
        'random-input-zones': (('drag-mining-input', (1, 7)),),
        'fixed-input-zones': (('drag-silo-input', (3, 17)),),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 18)))},
    "Challenge: KOHCTPYKTOP": {
        'obstructed': {
            # Middle rock
            *product(range(10, 13), range(4, 8)),
        },
        'fixed-input-zones': (('drag-mining-input', (3, 3)),
                              ('drag-silo-input', (2, 10)),
                              ('drag-silo-input', (2, 17))),
        'output-zones': (('output', (23, 2)),
                         ('output', (23, 7)),
                         ('output', (23, 12))),
        'recycler': ('drag-recycler', (26, 17))},
    "Ω-Pseudoethyne": {'obstructed': {}},  # Flidais components can't be added/modified so terrain may be empty
    "Σ-Ethylene": {'obstructed': {}},
    "Teleporters": {
        'obstructed': {
            # Top wall
            *product(range(0, 32), (0,)),
            *product(range(0, 8), (1,)),
            # Raised platform's edges
            *product(range(9, 11), range(1, 8)),
            *product(range(9, 32), range(8, 10)),
            # Right end of platform
            *product(range(28, 32), range(1, 8)),
            # TODO: Very hacky
            # Due to our simplifying assumptions aboout component pipe positions, we made drag-spaceship-input one row
            # shorter than it should be to handle its pipe being in its top row
            # Add terrain where the rest of it should be
            *product(range(4, 6), (20,)),
        },
        'fixed-input-zones': (('drag-spaceship-input', (4, 18)),),
        'output-zones': (('output', (23, 3)),),
        'recycler': ('drag-recycler', (26, 17))},
    "Precursor Compounds": {
        'obstructed': {
            # Top wall
            *product(range(0, 32), (0,)),
            *product(range(0, 8), (1,)),
            # Raised platform's edges
            *product(range(9, 11), range(1, 8)),
            *product(range(9, 32), range(8, 10)),
            # Right end of platform
            *product(range(28, 32), range(1, 8)),
            # TODO: Very hacky
            # See above note about drag-spaceship-input
            *product(range(1, 3), (11,)),
            *product(range(4, 6), (21,)),
        },
        'random-input-zones': (('drag-spaceship-input', (1, 9)),),
        'fixed-input-zones': (('drag-spaceship-input', (4, 19)),),
        'output-zones': (('output', (23, 12)),
                         ('output', (23, 18)))},
}
